﻿using System;
using System.Threading;
using System.Threading.Tasks;

namespace CV3
{
    class Program
    {
        static void Tick(object o)
        {
            Console.WriteLine("It's {0} seconds from start of the timer.", (int)(DateTime.Now - (DateTime)o).TotalSeconds);
        }

        static void ClassicThread()
        {
            for (int i = 0; ; i++)
            {
                Console.WriteLine("Line {0} from classic thread.", i);
                Thread.Sleep(1500);
            }
        }

        static void ThreadWithParameter(object o)
        {
            int count = (int)o;

            for (int i = 0; i < count; i++)
            {
                Console.WriteLine("Line {0} from parameterized thread.", i);
                Thread.Sleep(1000);
            }
        }

        delegate int MyNewDelegate(int x, int y); // Can use System.Func<int, int, int> delegate instead

        static int PrimeNumbersCount(int from, int to)
        {
            int count = 0;

            Parallel.For(from, to, i =>
            {
                for (int j = 2; j * j <= i; j++)
                    if (i % j == 0)
                        return;

                //count++;
                Interlocked.Increment(ref count);
            });

            Console.WriteLine("This simple method takes sooooo long time.");
            Thread.Sleep(5000);
            return count;
        }

        static void Main(string[] args)
        {
            Mutex mutex = new Mutex(false, "MyFirstMutex");

            if (!mutex.WaitOne(0))
            {
                Console.WriteLine("We're sorry but somebody decided that you can run only one instance of this application.");
                return;
            }

            // Timer
            Console.WriteLine("Start of the timer: {0}", DateTime.Now);
            Timer timer = new Timer(Tick, DateTime.Now, 2000, 1000);

            // Classic thread
            Console.WriteLine("Classic thread starts its job now.");
            Thread classic = new Thread(ClassicThread);
            classic.IsBackground = true;
            classic.Start();

            // Parameterized thread
            Thread parameterized = new Thread(ThreadWithParameter);
            parameterized.Start(10);
            Console.WriteLine("Now, main thread is working it's own job.");

            for (int i = 0; i < 5; i++)
            {
                Console.WriteLine("Line {0} from main thread.", i);
                Thread.Sleep(500);
            }

            Console.WriteLine("Main thread's job is finished and it'll wait for completition of the parameterized thread.");
            parameterized.Join();
            Console.WriteLine("Parameterized thread's job is finished now.");

            // Asynchronously delegate calling
            MyNewDelegate myDelegate = PrimeNumbersCount;
            IAsyncResult aResult = myDelegate.BeginInvoke(2, 185, null, null);
            Console.WriteLine("New thread is computing this and we can now compute something for ourselves.");
            int result = myDelegate.EndInvoke(aResult);
            Console.WriteLine("Computation finished, the answer is {0}.", result);

            while (Console.ReadKey().Key != ConsoleKey.Escape) ;

            mutex.ReleaseMutex();
        }
    }
}
